@@ -53,7 +53,7 @@ from __future__ import with_statement
 import os
 import os.path as op
 import heapq
-from itertools import izip_longest
+from itertools import izip_longest, count
 import csv
 import json
 from shutil import copy
@@ -216,7 +216,6 @@ class GeoBase(object):
         """
         # Main structure in which everything will be loaded
         # Dictionary of dictionary
-        self._data   = data
         self._things = {}
         self._ggrid  = None
 
@@ -229,6 +228,7 @@ class GeoBase(object):
         # This will be similar as _headers, but can be modified after loading
         # _headers is just for data loading
         self.fields = []
+        self.data   = data
 
         # Defaults
         props = {
@@ -286,6 +286,11 @@ class GeoBase(object):
         self._discard_dups  = props['discard_dups']
         self._verbose       = props['verbose']
 
+        # Some headers are not accepted
+        for h in self._headers:
+            if str(h).endswith('@raw') or str(h).startswith('__'):
+                raise ValueError('Header %s not accepted, should not end with "@raw" or start with "__".' % h)
+
         # Loading data
         self._configSubDelimiters()
 
@@ -301,6 +306,10 @@ class GeoBase(object):
             if self._verbose:
                 print 'Source was None, skipping loading...'
 
+            # We add those default fields if user adds data with self.set
+            self.fields = ['__key__', '__dup__', '__par__', '__lno__', '__gar__']
+
+
         # Grid
         if self.hasGeoSupport():
             self.createGrid()
@@ -314,6 +323,7 @@ class GeoBase(object):
         """Some precomputation on subdelimiters.
         """
         for h in self._headers:
+
             # If not in conf, do not sub split
             if h not in self._subdelimiters:
                 self._subdelimiters[h] = None
@@ -411,10 +421,24 @@ class GeoBase(object):
         return _reader
 
 
+    def _buildDuplicatedKey(self, key, nb_dups):
+        """
+        When the key is already in base and we do not want to discard the row,
+        we have to compute a new key for this row.
+        We iterate until we find an available key
+        """
+        for n in count(nb_dups):
+            d_key = '%s@%s' % (key, n)
+
+            if d_key not in self._things:
+                return d_key
+
+
+
     def _loadFile(self, source_fl):
         """Load the file and feed the self._things.
 
-        :param verbose: display informations or not during runtime
+        :param source_fl: file-like input
         :raises: IOError, if the source cannot be read
         :raises: ValueError, if duplicates are found in the source
         """
@@ -470,7 +494,8 @@ class GeoBase(object):
             else:
                 if discard_dups is False:
                     # We compute a new key for the duplicate
-                    d_key = '%s@%s' % (key, 1 + len(self._things[key]['__dup__']))
+                    nb_dups = 1 + len(self._things[key]['__dup__'])
+                    d_key   = self._buildDuplicatedKey(key, nb_dups)
 
                     # We update the data with this info
                     row_data['__key__'] = d_key
@@ -483,7 +508,7 @@ class GeoBase(object):
 
                     if verbose:
                         print "/!\ [lno %s] %s is duplicated #%s, first found lno %s: creation of %s..." % \
-                                (line_nb, key, len(self._things[key]['__dup__']), self._things[key]['__lno__'], d_key)
+                                (line_nb, key, nb_dups, self._things[key]['__lno__'], d_key)
                 else:
                     if verbose:
                         print "/!\ [lno %s] %s is duplicated, first found lno %s: dropping line..." % \
@@ -713,24 +738,28 @@ class GeoBase(object):
         For example, if you want to know all airports in Paris.
 
         :param conditions: a list of (field, value) conditions
-        :param reverse:    we look keys where the field is *not* the particular value
+        :param reverse:    we look keys where the field is *not* the particular value. \
+                Note that this negation is done at the lower level, before combining \
+                conditions. So if you have two conditions with mode='and', expect \
+                results matching not condition 1 *and* not condition 2.
         :param force_str:  for the str() method before every test
         :param mode:       either 'or' or 'and', how to handle several conditions
         :param from_keys:  if given, we will look for results from this iterable of keys
-        :returns:          an iterator of matching keys
+        :returns:          an iterable of (v, key) where v is the number of matched \
+                condition
 
         >>> list(geo_a.getKeysWhere([('city_code', 'PAR')]))
-        ['ORY', 'TNF', 'CDG', 'BVA']
+        [(1, 'ORY'), (1, 'TNF'), (1, 'CDG'), (1, 'BVA')]
         >>> list(geo_o.getKeysWhere([('comment', '')], reverse=True))
         []
         >>> list(geo_o.getKeysWhere([('__dup__', '[]')]))
         []
         >>> len(list(geo_o.getKeysWhere([('__dup__', [])])))
-        7024
+        7016
         >>> len(list(geo_o.getKeysWhere([('__dup__', '[]')], force_str=True)))
-        7024
+        7016
         >>> len(list(geo_o.getKeysWhere([('__par__', [])], reverse=True))) # Counting duplicated keys
-        4431
+        4438
 
         Testing several conditions.
 
@@ -739,11 +768,11 @@ class GeoBase(object):
         >>> len(list(geo_o.getKeysWhere(c_1)))
         18
         >>> len(list(geo_o.getKeysWhere(c_2)))
-        91
+        92
         >>> len(list(geo_o.getKeysWhere(c_1 + c_2, mode='and')))
         2
         >>> len(list(geo_o.getKeysWhere(c_1 + c_2, mode='or')))
-        107
+        108
 
         This works too \o/.
 
@@ -772,13 +801,14 @@ class GeoBase(object):
         elif mode == 'or':
             pass_all = any
         else:
-            raise ValueError('"mode" argument must be in %s' % str(['and', 'or']))
+            raise ValueError('"mode" argument must be in %s, was %s' % (str(['and', 'or']), mode))
 
 
         for key in from_keys:
             try:
-                if pass_all(pass_one(self.get(key, f), v) for f, v in conditions):
-                    yield key
+                matches = [pass_one(self.get(key, f), v) for f, v in conditions]
+                if pass_all(matches):
+                    yield sum(matches), key
             except KeyError:
                 # This means from_keys parameters contained unknown keys
                 if self._verbose:
@@ -791,7 +821,7 @@ class GeoBase(object):
         >>> str(geo_t)
         '<GeoBases.GeoBaseModule.GeoBase(stations) object at 0x...>'
         """
-        return '<GeoBases.GeoBaseModule.GeoBase(%s) object at 0x...>' % self._data
+        return '<GeoBases.GeoBaseModule.GeoBase(%s) object at 0x...>' % self.data
 
 
     def __iter__(self):
@@ -900,6 +930,10 @@ class GeoBase(object):
         >>> # Nice, stations <= 5km
         >>> [geo_t.get(k, 'name') for d, k in sorted(geo_t.findNearPoint((43.70, 7.26), 5))]
         ['Nice-Ville', 'Nice-Riquier', 'Nice-St-Roch', 'Villefranche-sur-Mer', 'Nice-St-Augustin']
+        >>>
+        >>> # Wrong geocode
+        >>> sorted(geo_t.findNearPoint(None, 5))
+        []
 
         No grid mode.
 
@@ -945,7 +979,7 @@ class GeoBase(object):
         We just look up in the base to retrieve lat/lng, and
         call findNearPoint.
 
-        :param key:     the key
+        :param key:     the key of the thing (like 'SFO')
         :param radius:  the radius of the search (kilometers)
         :param from_keys: if None, it takes all keys in consideration, else takes from_keys \
             iterable of keys to perform search.
@@ -1000,10 +1034,6 @@ class GeoBase(object):
         we look for the closest thing from this point, given by
         latitude/longitude.
 
-        Note that a similar implementation is done in
-        the LocalHelper, to find efficiently N closest point
-        in a graph, from a point (using heaps).
-
         :param lat_lng:   the lat_lng of the point (a tuple of (lat, lng))
         :param N:         the N closest results wanted
         :param from_keys: if None, it takes all keys in consideration, else takes from_keys \
@@ -1024,7 +1054,8 @@ class GeoBase(object):
         >>> # Corner case, from_keys empty is not used
         >>> list(geo_t.findClosestFromPoint((43.70, 7.26), N=2, from_keys=()))
         []
-        >>> 
+        >>> list(geo_t.findClosestFromPoint(None, N=2))
+        []
         >>> #from datetime import datetime
         >>> #before = datetime.now()
         >>> #for _ in range(100): s = geo_a.findClosestFromPoint((43.70, 7.26), N=3)
@@ -1057,6 +1088,62 @@ class GeoBase(object):
                 yield (dist, thing)
 
 
+
+    def findClosestFromKey(self, key, N=1, from_keys=None, grid=True, double_check=True):
+        """
+        Same as findClosestFromPoint, except the point is given
+        not by a lat/lng, but with its key, like ORY or SFO.
+        We just look up in the base to retrieve lat/lng, and
+        call findClosestFromPoint.
+
+        :param key:       the key of the thing (like 'SFO')
+        :param N:         the N closest results wanted
+        :param from_keys: if None, it takes all keys in consideration, else takes from_keys \
+            iterable of keys to perform findClosestFromPoint. This is useful when we have names \
+            and have to perform a matching based on name and location (see fuzzyGetAroundLatLng).
+        :param grid:    boolean, use grid or not
+        :param double_check: when using grid, perform an additional check on results distance, \
+            this is useful because the grid is approximate, so the results are only as accurate \
+            as the grid size
+        :returns:       an iterable of (distance, key) like [(3.2, 'SFO'), (4.5, 'LAX')]
+
+        >>> list(geo_a.findClosestFromKey('ORY')) # Orly
+        [(0.0, 'ORY')]
+        >>> list(geo_a.findClosestFromKey('ORY', N=3))
+        [(0.0, 'ORY'), (18.80..., 'TNF'), (27.80..., 'LBG')]
+        >>> # Corner case, from_keys empty is not used
+        >>> list(geo_t.findClosestFromKey('ORY', N=2, from_keys=()))
+        []
+        >>> list(geo_t.findClosestFromKey(None, N=2))
+        []
+        >>> #from datetime import datetime
+        >>> #before = datetime.now()
+        >>> #for _ in range(100): s = geo_a.findClosestFromKey('NCE', N=3)
+        >>> #print(datetime.now() - before)
+
+        No grid.
+
+        >>> list(geo_o.findClosestFromKey('ORY', grid=False)) # Nice
+        [(0.0, 'ORY')]
+        >>> list(geo_a.findClosestFromKey('ORY', N=3, grid=False)) # Nice
+        [(0.0, 'ORY'), (18.80..., 'TNF'), (27.80..., 'LBG')]
+        >>> list(geo_t.findClosestFromKey('frnic', N=1, grid=False)) # Nice
+        [(0.0, 'frnic')]
+        >>> list(geo_t.findClosestFromKey('frnic', N=2, grid=False, from_keys=('frpaz', 'frply', 'frbve')))
+        [(482.79..., 'frbve'), (683.52..., 'frpaz')]
+        """
+        if from_keys is None:
+            from_keys = iter(self)
+
+        if grid:
+            for dist, thing in self._ggrid.findClosestFromKey(key, N, double_check, from_keys):
+                yield (dist, thing)
+
+        else:
+            for dist, thing in self.findClosestFromPoint(self.getLocation(key), N, from_keys, grid, double_check):
+                yield (dist, thing)
+
+
     def _buildRatios(self, fuzzy_value, field, keys, min_match=0):
         """
         Compute the iterable of (dist, keys) of a reference
@@ -1192,6 +1279,7 @@ class GeoBase(object):
                        field,
                        max_results=None,
                        min_match=0.75,
+                       from_keys=None,
                        verbose=True,
                        show_bad=(1, 1)):
         """
@@ -1201,6 +1289,8 @@ class GeoBase(object):
         :param field:       the field we look into, like 'name'
         :param max_results: if None, returns all, if an int, only returns the first ones
         :param min_match:   filter out matches under this threshold
+        :param from_keys:   if None, it takes all keys into consideration, else takes from_keys \
+            iterable of keys as search domain
         :param verbose:     display information on a certain range of similarity
         :param show_bad:    the range of similarity
         :returns:           an iterable of (distance, key) like [(0.97, 'SFO'), (0.55, 'LAX')]
@@ -1217,16 +1307,16 @@ class GeoBase(object):
 
         Some biasing:
 
-        >>> geo_a.biasFuzzyCache('paris de gaulle', 'name', None, 0.75, [(0.5, 'Biased result')])
+        >>> geo_a.biasFuzzyCache('paris de gaulle', 'name', None, 0.75, None, [(0.5, 'Biased result')])
         >>> geo_a.fuzzyGetCached('paris de gaulle', 'name', max_results=None, show_bad=(0, 1))[0] # Cache there
         (0.78..., 'CDG')
         >>> geo_a.clearCache()
         >>> geo_a.fuzzyGetCached('paris de gaulle', 'name', max_results=None, min_match=0.75)
-        Using bias: ('paris+de+gaulle', 'name', None, 0.75)
+        Using bias: ('paris+de+gaulle', 'name', None, 0.75, None)
         [(0.5, 'Biased result')]
         """
         # Cleaning is for keeping only useful data
-        entry = self._buildCacheKey(fuzzy_value, field, max_results, min_match)
+        entry = self._buildCacheKey(fuzzy_value, field, max_results, min_match, from_keys)
 
         if entry not in self._cache_fuzzy:
 
@@ -1242,7 +1332,7 @@ class GeoBase(object):
 
 
 
-    def biasFuzzyCache(self, fuzzy_value, field, max_results, min_match, biased_result):
+    def biasFuzzyCache(self, fuzzy_value, field, max_results, min_match, from_keys, biased_result):
         """
         If algorithms for fuzzy searches are failing on a single example,
         it is possible to use a first cache which will block
@@ -1252,12 +1342,14 @@ class GeoBase(object):
         :param field:         the field we look into, like 'name'
         :param max_results:   if None, returns all, if an int, only returns the first ones
         :param min_match:     filter out matches under this threshold
+        :param from_keys:     if None, it takes all keys into consideration, else takes from_keys \
+            iterable of keys as search domain
         :param biased_result: the expected result
         :returns:             None
 
         """
         # Cleaning is for keeping only useful data
-        entry = self._buildCacheKey(fuzzy_value, field, max_results, min_match)
+        entry = self._buildCacheKey(fuzzy_value, field, max_results, min_match, from_keys)
 
         self._bias_cache_fuzzy[entry] = biased_result
 
@@ -1275,15 +1367,15 @@ class GeoBase(object):
 
 
     @staticmethod
-    def _buildCacheKey(fuzzy_value, field, max_results, min_match):
+    def _buildCacheKey(fuzzy_value, field, max_results, min_match, from_keys):
         """Key for the cache of fuzzyGet, based on parameters.
 
-        >>> geo_a._buildCacheKey('paris de gaulle', 'name', max_results=None, min_match=0)
-        ('paris+de+gaulle', 'name', None, 0)
-        >>> geo_a._buildCacheKey('Antibes SNCF 2', 'name', max_results=3, min_match=0)
-        ('antibes', 'name', 3, 0)
+        >>> geo_a._buildCacheKey('paris de gaulle', 'name', max_results=None, min_match=0, from_keys=None)
+        ('paris+de+gaulle', 'name', None, 0, None)
+        >>> geo_a._buildCacheKey('Antibes SNCF 2', 'name', max_results=3, min_match=0, from_keys=None)
+        ('antibes', 'name', 3, 0, None)
         """
-        return '+'.join(clean(fuzzy_value)), field, max_results, min_match
+        return '+'.join(clean(fuzzy_value)), field, max_results, min_match, from_keys
 
 
     def _debugFuzzy(self, match, fuzzy_value, field, show_bad=(1, 1)):
@@ -1340,7 +1432,13 @@ class GeoBase(object):
         # If the key is not in the base,
         # we simply add it
         if key not in self._things:
-            self._things[key] = {}
+            self._things[key] = {
+                '__key__' : key,      # special field for key
+                '__lno__' : 0,        # special field for line number
+                '__gar__' : [],       # special field for garbage
+                '__dup__' : [],       # special field for duplicates
+                '__par__' : [],       # special field for parent
+            }
 
         self._things[key][field] = value
 
@@ -1547,8 +1645,10 @@ class GeoBase(object):
         if add_lines is None:
             add_lines = []
 
+        dup_lines = []
+
         if link_duplicates:
-            # We add to add_lines all list of duplicates
+            # We add to dup_lines all list of duplicates
             # We keep a set of already processed "master" keys to avoid
             # putting several identical lists in the json
             done_keys = set()
@@ -1563,11 +1663,11 @@ class GeoBase(object):
 
                 if self.hasDuplicates(key) and not mkey.issubset(done_keys):
                     # mkey have some keys which are not in done_keys
-                    add_lines.append(self.getAllDuplicates(key, '__key__'))
+                    dup_lines.append(self.getAllDuplicates(key, '__key__'))
                     done_keys = done_keys | mkey
 
             if verbose:
-                print '* Added lines for duplicates linking, total %s' % len(add_lines)
+                print '* Added lines for duplicates linking, total %s' % len(dup_lines)
 
         # Count the categories for coloring
         categories = {}
@@ -1678,7 +1778,7 @@ class GeoBase(object):
         # Gathering data for lines
         data_lines = []
 
-        for line in add_lines:
+        for line in add_lines + dup_lines:
             data_line = []
 
             for l_key in line:
@@ -1709,6 +1809,7 @@ class GeoBase(object):
                     'icon_type'       : icon_type,
                     'base_icon'       : base_icon,
                     'link_duplicates' : link_duplicates,
+                    'nb_user_lines'   : len(add_lines),
                 },
                 'points'     : data,
                 'lines'      : data_lines,
@@ -1762,6 +1863,8 @@ def ext_split(value, split):
     :param split:  the splitter
     :returns:      the split value
 
+    >>> ext_split('', ',')
+    ()
     >>> ext_split('PAR', 'A')
     ('P', 'R')
     >>> ext_split('PAR', '')
@@ -1775,6 +1878,11 @@ def ext_split(value, split):
         # Here we convert a string like 'CA' into ('C', 'A')
         return tuple(value)
 
+    # Python split function has ''.split(';') -> ['']
+    # But in this case we prefer having [] as a result
+    if not value:
+        return ()
+
     return tuple(value.split(split))
 
 
@@ -1787,6 +1895,17 @@ def recursive_split(value, splits):
 
     >>> recursive_split('PAR^Paris/Parys', ['^', '/'])
     (('PAR',), ('Paris', 'Parys'))
+    >>> recursive_split('|PAR|=', ['=', '|'])
+    (('', 'PAR', ''),)
+
+    Multiple splits on empty string should return empty tuple.
+
+    >>> recursive_split('', ['^'])
+    ()
+    >>> recursive_split('', ['^', '/'])
+    ()
+    >>> recursive_split('', ['^', '/', ':'])
+    ()
     """
     # Case where no subdelimiters
     if not splits:
@@ -1796,13 +1915,13 @@ def recursive_split(value, splits):
         return ext_split(value, splits[0])
 
     if len(splits) == 2:
-        return tuple(ext_split(v, splits[1]) for v in value.split(splits[0]))
+        return tuple(ext_split(v, splits[1])
+                     for v in value.split(splits[0]) if v)
 
     if len(splits) == 3:
-        return tuple(
-            tuple(ext_split(sv, splits[2]) for sv in ext_split(v, splits[1]))
-            for v in value.split(splits[0])
-        )
+        return tuple(tuple(ext_split(sv, splits[2])
+                           for sv in ext_split(v, splits[1]) if sv)
+                     for v in value.split(splits[0]) if v)
 
     raise ValueError('Sub delimiter "%s" not supported.' % str(splits))
 
